/*
Astrolabe

Copyright (C) 2013 Observatoire de Paris-Meudon

Ce programme est un logiciel libre ; vous pouvez le redistribuer et/ou le modifier conformément aux dispositions de la Licence Publique Générale GNU, telle que publiée par la Free Software Foundation ; version 2 de la licence, ou encore (à votre choix) toute version ultérieure.

Ce programme est distribué dans l'espoir qu'il sera utile, mais SANS AUCUNE GARANTIE ; sans même la garantie implicite de COMMERCIALISATION ou D'ADAPTATION A UN OBJET PARTICULIER. Pour plus de détail, voir la Licence Publique Générale GNU .

Vous devez avoir reçu un exemplaire de la Licence Publique Générale GNU en même temps que ce programme ; si ce n'est pas le cas, écrivez à la Free Software Foundation Inc., 675 Mass Ave, Cambridge, MA 02139, Etats-Unis.
*/

/**
 * Simulation d'un astrolabe
 */

function isCanvasSupported() {
    var elem = document.createElement('canvas');
    return !!(elem.getContext && elem.getContext('2d'));
}

function versionNavigateur() {
    document.body.appendChild(document.createTextNode("Cette simulation nécessite un navigateur récent."));
}

if (isCanvasSupported()) {
    new Astrolabe();
} else {
    if (window.addEventListener)
        window.addEventListener('load', function() { versionNavigateur(); }, false);
    else if (window.attachEvent)
        window.attachEvent('onload', function() { versionNavigateur(); });
    
}

function Astrolabe() {
    var date, params, p, i;
    
    this.applet = new SimuApplet(this);
    this.applet.appletDoc('astrolabe_doc.xml');
    
    // pour changer la valeur par défaut de la date à partir de la date du jour
    this.applet.lireParametres();
    date = new Date();
    params = this.applet.getInParams();
    for (i=0; i<params.length; i++) {
        p = params[i];
        if (p.label == "jour")
            p.defaut = "" + date.getDate();
        else if (p.label == "mois")
            p.defaut = "" + (date.getMonth() + 1);
    }
    
    this.applet.setDelaiIterations(150);
    this.n = 192;
    
    this.phi = 48.8;
    this.jour = 1;
    this.mois = 1;
    
    this.canvas = null;
    this.ctx = null;
}

Astrolabe.prototype.initCalculLive = function(lin) {
    var affImage, obli, l, ads;
    
    this.phi = lin.lireDouble('phi') * Math.PI / 180;
    if (this.phi < 0)
        throw "La simulation ne gère que les latitudes Nord";
    this.j = lin.lireDouble('jour');
    this.m = lin.lireDouble('mois');
    this.opt_azimuts = ("on" == lin.lireString('optazimuts'));
    this.opt_pts_cardinaux = ("on" == lin.lireString('optptscardinaux'));
    this.opt_declinaisons = ("on" == lin.lireString('optdeclinaisons'));
    this.opt_ecliptique = ("on" == lin.lireString('optecliptique'));
    this.opt_horizon = ("on" == lin.lireString('opthorizon'));
    this.opt_mois = ("on" == lin.lireString('optmois'));
    this.opt_etoiles = ("on" == lin.lireString('optetoiles'));
    this.opt_constellations = ("on" == lin.lireString('optconstellations'));
    //this.opt_crepuscule = ("on" == lin.lireString('optcrepuscule'));
    
    this.etoiles = this.constructionEtoiles(this.lectureFichier("stars.txt"));
    this.constellations = this.constructionConstellations(this.lectureFichier("constellations.txt"));
    
    affImage = this.applet.getAffichageChoisi();
    this.imgdimx = affImage.imgdimx;
    this.imgdimy = affImage.imgdimy;
    this.rMax = 1.65;
    this.dr = (Math.min(this.imgdimx, this.imgdimy) / 2) / this.rMax;
    this.dx = this.imgdimx / 2;
    this.dy = this.imgdimy / 2;
    this.itps = 0;
    this.tps = 0;
    this.obli = 23.5 * Math.PI / 180.; // obliquité
    obli = this.obli;
    
    this.j += nbjours(this.m);
    this.l = (this.j - 81.) * 360. / 365.25 * Math.PI/180.; // long. ecli. du soleil
    l = this.l;
    
    ads = (Math.cos(l) > 0 ?
        Math.atan(Math.tan(l) * Math.cos(obli)) :
        Math.PI + Math.atan(Math.tan(l) * Math.cos(obli)));
    // ascention droite du soleil
    ads = (ads < -Math.PI ? ads + 2*Math.PI : (ads > Math.PI ? ads - 2*Math.PI : ads));
    this.ads = ads;

    this.canvas = document.createElement('canvas');
    this.canvas.setAttribute('width', this.imgdimx);
    this.canvas.setAttribute('height', this.imgdimy);
    this.ctx = this.canvas.getContext('2d');
}

Astrolabe.prototype.calculLive = function() {
    var out, phi, l, ads, obli, ts, hsv, alpha, delta, z, x0, y0, r, A, alphas, deltas,
        i, etoile, constellation, etoile2, x, y;
    
    phi = this.phi;
    l = this.l;
    ads = this.ads;
    obli = this.obli;
    
    this.ctx.fillStyle = 'white';
    this.ctx.fillRect(0, 0, this.imgdimx-1, this.imgdimy-1);
    
    this.ctx.beginPath();
    this.ctx.arc(this.dx, this.dy, this.dr * this.rMax, 0, 2*Math.PI, false);
    this.ctx.clip();
    
    ts = this.tps; // tps sidéral (entre 0 et 24h)
    ts = (ts > 24. ?  ts - 24 : (ts < 0 ? ts + 24. : ts));
    ts = ts * Math.PI / 12.;
    this.ts = ts;

    hsv = ts - ads + Math.PI; // heure solaire vraie (locale)
    hsv = (hsv > 2.*Math.PI ? hsv - 2.*Math.PI : hsv);
    this.hsv = hsv;
    
    this.disque(0, 0, this.rMax, '#90E0A0');
    
    alpha = 0.0;
    delta = phi;
    // crépuscule
    /*
    if (this.opt_crepuscule) {
        z = 96. * Math.PI/180.;
        x0 = xcp(alpha, delta, z);
        y0 = ycp(alpha, delta, z);
        r = rayon(alpha, delta, z);
        this.disque(x0, y0, r, '#B0B0B0');
    }
    */
    
    // horizon
    if (this.opt_horizon) {
        z = 90. * Math.PI/180.;
        x0 = xcp(alpha, delta, z);
        y0 = ycp(alpha, delta, z);
        r = rayon(alpha, delta, z);
        this.disque(x0, y0, r, leve(phi, l, hsv, obli) == 1 ? '#C0F0FF' : '#303070');
        this.cercle(x0, y0, r, '#30A050', 1);
    }
    
    // zenith
    if (this.opt_azimuts) {
        alpha = 0.0;
        delta = phi;
        this.point(xp(alpha, delta), yp(alpha, delta), '#000000', 4);
    }
    
    // azimuth
    if (this.opt_azimuts) {
        z = 90. * Math.PI/180.;
        this.ligne(-this.rMax, 0, this.rMax, 0, '#000000', 1); // N-S
        A = 0.; //E-O
        delta = deltap(phi, A);
        alpha = alphap(phi, A, delta);
        x0 = xcp(alpha, delta, z);
        y0 = ycp(alpha, delta, z);
        r = rayon(alpha, delta, z);
        this.cercle(x0, y0, r, '#000000', 1);
        A = -45.*Math.PI/180.; //SO-NE
        delta = deltap(phi, A);
        alpha = alphap(phi, A, delta);
        x0 = xcp(alpha, delta, z);
        y0 = ycp(alpha, delta, z);
        r = rayon(alpha, delta, z);
        this.cercle(x0, y0, r, '#000000', 1);
        A = 45.*Math.PI/180.; //SE-NO
        delta = deltap(phi, A);
        alpha = alphap(phi, A, delta);
        x0 = xcp(alpha, delta, z);
        y0 = ycp(alpha, delta, z);
        r = rayon(alpha, delta, z);
        this.cercle(x0, y0, r, '#000000', 1);
        A = 90.*Math.PI/180.; //S-N
        delta = deltap(phi, A);
        alpha = alphap(phi, A, delta);
        x0 = xcp(alpha, delta, z);
        y0 = ycp(alpha, delta, z);
        r = rayon(alpha, delta, z);
        this.cercle(x0, y0, r, '#000000', 1);
    }
    
    if (this.opt_pts_cardinaux) {
        // points cardinaux
        A = 0.;
        delta = deltap(phi, A);
        alpha = alphap(phi, A, delta);
        this.texte('S', xp(alpha, delta)*1.08, yp(alpha, delta)*1.08, '#000000');
        A = Math.PI / 4;
        delta = deltap(phi, A);
        alpha = alphap(phi, A, delta);
        this.texte('SO', xp(alpha, delta)*1.08, yp(alpha, delta)*1.08, '#000000');
        A = Math.PI / 2;
        delta = deltap(phi, A);
        alpha = alphap(phi, A, delta);
        this.texte('O', xp(alpha, delta)*1.08, yp(alpha, delta)*1.08, '#000000');
        A = Math.PI * 3/4;
        delta = deltap(phi, A);
        alpha = alphap(phi, A, delta);
        this.texte('NO', xp(alpha, delta)*1.08, yp(alpha, delta)*1.08, '#000000');
        A = Math.PI;
        delta = deltap(phi, A);
        alpha = alphap(phi, A, delta);
        this.texte('N', xp(alpha, delta)*1.08, yp(alpha, delta)*1.08, '#000000');
        A = Math.PI * 5/4;
        delta = deltap(phi, A);
        alpha = alphap(phi, A, delta);
        this.texte('NE', xp(alpha, delta)*1.08, yp(alpha, delta)*1.08, '#000000');
        A = -Math.PI / 2;
        delta = deltap(phi, A);
        alpha = alphap(phi, A, delta);
        this.texte('E', xp(alpha, delta)*1.08, yp(alpha, delta)*1.08, '#000000');
        A = -Math.PI / 4;
        delta = deltap(phi, A);
        alpha = alphap(phi, A, delta);
        this.texte('SE', xp(alpha, delta)*1.08, yp(alpha, delta)*1.08, '#000000');
    }
    
    // projection de l'araignée :
    // (indépendant de la latitude du lieu, mais dépend du tps sauf pour cercles de déclinaison)

    // projection des cercles de declinaison (indpt du tps)
    
    // equateur
    if (this.opt_declinaisons) {
        delta = 0.0;
        r = Math.abs(Math.tan(Math.PI * 0.25 - delta * 0.5));
        this.cercle(0, 0, r, '#FF0000', 3);
        //this.texte('δ=0°', xp(Math.PI/3, 0)*1.1, yp(Math.PI/3, 0)*1.1, '#FF0000');
        
        // cercle antartique
        delta = -Math.PI * 0.5 + obli;
        r = Math.abs(Math.tan(Math.PI * 0.25 - delta * 0.5));
        this.cercle(0, 0, r, '#FF0000', 1);
        
        // tropique capricorne
        delta = -obli;
        r = Math.abs(Math.tan(Math.PI * 0.25 - delta * 0.5));
        this.cercle(0, 0, r, '#FF0000', 1);
        
        // tropique cancer
        delta = obli;
        r = Math.abs(Math.tan(Math.PI * 0.25 - delta * 0.5));
        this.cercle(0, 0, r, '#FF0000', 1);
        
        // cercle artique
        delta = Math.PI * 0.5 - obli;
        r = Math.abs(Math.tan(Math.PI * 0.25 - delta * 0.5));
        this.cercle(0, 0, r, '#FF0000', 1);
    }
    
    // étoiles
    if (this.opt_etoiles) {
        for (i=0; i<this.etoiles.length; i++) {
            etoile = this.etoiles[i];
            alpha = etoile.alpha;
            delta = etoile.delta;
            this.point(xp((ts - alpha), delta),
                yp((ts - alpha), delta), '#FFA000', 2);
        }
    }
    // constellations
    if (this.opt_constellations) {
        this.ctx.strokeStyle = '#FFA000';
        this.ctx.lineWidth = 1;
        for (i=0; i<this.constellations.length; i++) {
            this.ctx.beginPath();
            constellation = this.constellations[i];
            etoile2 = null;
            for (j=0; j<constellation.length; j++) {
                etoile = constellation[j];
                alpha = etoile.alpha;
                delta = etoile.delta;
                x = xp((ts - alpha), delta);
                x = this.dx + x * this.dr;
                y = yp((ts - alpha), delta);
                y = this.dy - y * this.dr;
                if (j == 0)
                    this.ctx.moveTo(x, y);
                else
                    this.ctx.lineTo(x, y);
            }
            this.ctx.stroke();
        }
    }
    
    // projection cercle de l'écliptique
    if (this.opt_ecliptique) {
        alpha = ts + Math.PI * 0.5;
        delta = Math.PI * 0.5 - obli;
        z = Math.PI * 0.5;
        x0 = xcp(alpha, delta, z);
        y0 = ycp(alpha, delta, z);
        r = rayon(alpha, delta, z);
        this.cercle(x0, y0, r, '#FFFF00', 3);
    }
    
    // le point vernal
    if (this.opt_declinaisons && this.opt_ecliptique) {
        alpha = ts;
        delta = 0.;
        this.point(xp(alpha, delta), yp(alpha, delta), '#FF0000', 4);
        this.texte('γ', xp(alpha, delta)*1.1, yp(alpha, delta)*1.1, '#FF0000');
    }
    
    // le Soleil
    delta = Math.PI * 0.5 - obli;
    alpha = ts + Math.PI * 0.5;
    if (this.opt_ecliptique) {
        A = l - Math.PI * 0.5;
        deltas = Math.asin(Math.cos(delta) * Math.cos(A));
        alphas = Math.asin(Math.sin(A) / Math.cos(deltas));
        alphas = (-Math.sin(delta) * Math.cos(A) < 0 ? Math.PI - alphas : alphas);
        this.point(xp(alpha + alphas, deltas), yp(alpha + alphas, deltas),
            leve(phi, l, hsv, obli) == 1 ? '#FFFF00' : '#000000', 6);
    }
    
    // le premier de chaque mois
    if (this.opt_mois) {
        for (j=1; j<=12; j++)
            this.premier(j, alpha, delta);
    }
   
    out = new ListeValeurs();
    out.ajouterString('infos', this.infos());
    out.ajouter('image', this.canvas);
    this.itps++;
    if (this.itps > this.n)
        this.itps -= this.n;
    this.tps = (this.itps * 24) / this.n;
    return out;
}

Astrolabe.prototype.premier = function(j, alpha, delta) {
    var nj, l, A, alphas, deltas, z;
    
    nj = 1 + nbjours(j);
    l = (nj - 81.) * 360./365.25 * Math.PI/180.; // long. ecli. du 1/1
    A = l - Math.PI * 0.5;
    deltas = Math.asin(Math.cos(delta) * Math.cos(A));
    alphas = Math.asin(Math.sin(A) / Math.cos(deltas));
    alphas = (-Math.sin(delta) * Math.cos(A) < 0 ? alphas = Math.PI - alphas : alphas);
    z = Math.PI * 0.5;
    this.texte(j, 1.1 * xp(alpha + alphas, deltas) - 0.1 * xcp(alpha, delta, z),
        1.1 * yp(alpha + alphas, deltas) - 0.1 * ycp(alpha, delta, z), '#0000FF');
    this.point(xp(alpha + alphas, deltas), yp(alpha + alphas, deltas), '#0000FF', 3);
}

Astrolabe.prototype.infos = function() {
    // "ts=" + (this.ts * 12. / Math.PI).toFixed(3) + " h"
    return("heure solaire vraie=" + (this.hsv * 12. / Math.PI).toFixed(3)) + " h";
}

Astrolabe.prototype.cercle = function(x, y, r, couleur, epaisseur) {
    this.ctx.strokeStyle = couleur;
    this.ctx.lineWidth = epaisseur;
    this.ctx.beginPath();
    this.ctx.arc(this.dx + x * this.dr, this.dy - y * this.dr, r * this.dr, 0, 2*Math.PI, false);
    this.ctx.stroke();
}

Astrolabe.prototype.disque = function(x, y, r, couleur) {
    this.ctx.fillStyle = couleur;
    this.ctx.beginPath();
    this.ctx.arc(this.dx + x * this.dr, this.dy - y * this.dr, r * this.dr, 0, 2*Math.PI, false);
    this.ctx.fill();
}

Astrolabe.prototype.point = function(x, y, couleur, rayon) {
    this.ctx.fillStyle = couleur;
    this.ctx.beginPath();
    this.ctx.arc(this.dx + x * this.dr, this.dy - y * this.dr, rayon, 0, 2*Math.PI, false);
    this.ctx.fill();
}

Astrolabe.prototype.ligne = function(x1, y1, x2, y2, couleur, epaisseur) {
    this.ctx.strokeStyle = couleur;
    this.ctx.lineWidth = epaisseur;
    this.ctx.beginPath();
    this.ctx.moveTo(this.dx + x1 * this.dr, this.dy - y1 * this.dr);
    this.ctx.lineTo(this.dx + x2 * this.dr, this.dy - y2 * this.dr);
    this.ctx.stroke();
}

Astrolabe.prototype.texte = function(s, x, y, couleur) {
    this.ctx.font = '10px sans-serif';
    this.ctx.fillStyle = couleur;
    var metrics = this.ctx.measureText(s);
    this.ctx.fillText(s, this.dx + x * this.dr - metrics.width / 2, this.dy - y * this.dr + 5);
}

Astrolabe.prototype.constructionEtoiles = function(lignes) {
    var etoiles, i, ligne, alpha, delta;
    
    etoiles = new Array();
    for (i=0; i<lignes.length; i++) {
        ligne = lignes[i];
        if (ligne.length >= 2) {
            alpha = parseFloat(ligne[0]);
            delta = parseFloat(ligne[1]);
            etoiles.push(new Etoile(alpha * Math.PI/12., delta * Math.PI/180.));
        }
    }
    return(etoiles);
}

Astrolabe.prototype.constructionConstellations = function(lignes) {
    var constellations, etoiles, alpha, delta;
    
    // en fait une "constellation" ici représente un sous-ensemble d'une constellation,
    // tracé par un ensemble d'étoiles
    constellations = new Array();
    etoiles = null;
    for (i=0; i<lignes.length; i++) {
        ligne = lignes[i];
        if (ligne.length == 2 && ligne[0].charAt(0) != '#') {
            if (etoiles == null)
                etoiles = new Array();
            alpha = parseFloat(ligne[0]);
            delta = parseFloat(ligne[1]);
            etoiles.push(new Etoile(alpha * Math.PI/12., delta * Math.PI/180.));
        } else if (etoiles != null) {
            constellations.push(etoiles);
            etoiles = null;
        }
    }
    return(constellations);
}

/**
 * Lecture d'un fichier ASCII utilisant des espaces et retours à la ligne comme séparateurs
 * Renvoie un tableau des lignes, chaque ligne étant un tableau de strings
 */
Astrolabe.prototype.lectureFichier = function(nomFichier) {
    var xmlhttp, slignes, tlignes, i, ligne, mots, ind1, j;
    
    xmlhttp = new XMLHttpRequest();
    if (xmlhttp.overrideMimeType)
        xmlhttp.overrideMimeType('text/plain');
    
    try {
        xmlhttp.open("GET", nomFichier, false);
        xmlhttp.send();
        slignes = xmlhttp.responseText.split("\n");
        tlignes = new Array(slignes.length);
        for (i=0; i<slignes.length; i++) {
            ligne = slignes[i];
            mots = new Array();
            
            ind1 = -1;
            for (j=0; j<ligne.length; j++) {
                if (ligne[j] != ' ' && ligne[j] != '\t') {
                    if (ind1 == -1)
                        ind1 = j;
                } else {
                    if (ind1 != -1) {
                        mots.push(ligne.substring(ind1, j));
                        ind1 = -1;
                    }
                }
            }
            if (ind1 != -1)
                mots.push(ligne.substring(ind1));
            
            tlignes[i] = mots;
        }
        return(tlignes);
    } catch (ex) {
        alert("Erreur à la lecture du fichier de données: " + ex);
        return(null);
    }
}

function Etoile(alpha, delta) {
    this.alpha = alpha;
    this.delta = delta;
}

function Constellation(tetoiles) {
    // tetoiles : tableau de tableaux d'étoiles
    this.tetoiles = tetoiles;
}

function nbjours(mois) {
    var nj;
    nj = 0;
    if (mois > 1)
        nj += 31;
    if (mois > 2)
        nj += 28;
    if (mois > 3)
        nj += 31;
    if (mois > 4)
        nj += 30;
    if (mois > 5)
        nj += 31;
    if (mois > 6)
        nj += 30;
    if (mois > 7)
        nj += 31;
    if (mois > 8)
        nj += 31;
    if (mois > 9)
        nj += 30;
    if (mois > 10)
        nj += 31;
    if (mois > 11)
        nj += 30;
    return(nj);
}

function xp(alpha, delta) {
    // projection stereographique
    return(Math.tan(Math.PI * 0.25 - delta * 0.5) * Math.cos(alpha));
}

function yp(alpha, delta) {
    return(Math.tan(Math.PI * 0.25 - delta * 0.5) * Math.sin(alpha));
}

function dist(x1, y1, x2, y2) {
    return(Math.sqrt(Math.pow(x1 - x2, 2) + Math.pow(y1 - y2, 2)));
}

// les fonctions suivantes ne servent qu'à définir le rayon et le centre de la projection
// d'un cercle de la sphère céleste de centre E(alpha,delta) et de rayon z
function a1(alpha, delta, z) {
    return((delta + z < Math.PI * 0.5 ? alpha : Math.PI + alpha));
}

function b1(alpha, delta, z) {
    return((delta + z < Math.PI * 0.5 ? delta + z : Math.PI - (delta + z)));
}

function xp1(alpha, delta, z) {
    return(xp(a1(alpha, delta, z), b1(alpha, delta, z)));
}

function yp1(alpha, delta, z) {
    return(yp(a1(alpha, delta, z), b1(alpha, delta, z)));
}

function a2(alpha, delta, z) {
    return((delta - z > -Math.PI * 0.5 ? alpha : Math.PI + alpha));
}

function b2(alpha, delta, z) {
    return((delta-z > -Math.PI * 0.5 ? delta - z : -Math.PI - (delta - z)));
}

function xp2(alpha, delta, z) {
    return(xp(a2(alpha, delta, z), b2(alpha, delta, z)));
}

function yp2(alpha, delta, z) {
    return(yp(a2(alpha, delta, z), b2(alpha, delta, z)));
}

function rayon(alpha, delta, z) {
    // le rayon
    return(dist(xp1(alpha, delta, z), yp1(alpha, delta, z), xp2(alpha, delta, z), yp2(alpha, delta, z)) * 0.5);
}

function xcp(alpha, delta, z) {
    // coordonnées du centre
    return((xp1(alpha, delta, z) + xp2(alpha, delta, z)) * 0.5);
}

function ycp(alpha, delta, z) {
    return((yp1(alpha, delta, z) + yp2(alpha, delta, z)) * 0.5);
}

// fonctions utiles pour tracer des méridien à partir de la définition
// du pole du méridien (qui correspond alors à un cercle sur la sphère céleste de rayon 90°)

function deltap(phi, A) {
    return(Math.asin(-Math.cos(phi) * Math.cos(A)));
}

function alphap(phi, A, delta) {
    return((Math.sin(phi) * Math.cos(A) > 0. ? Math.asin(Math.sin(A) / Math.cos(delta)) :
        Math.PI - Math.asin(Math.sin(A) / Math.cos(delta))));
}

// détermine si le soleil est couché ou levé
function delta_sol(lsol, obli) {
    return(Math.asin(Math.cos(Math.PI * 0.5 - obli) * Math.cos(lsol - Math.PI * 0.5)));
}

function sinush(phi, lsol, hsv, obli) {
    var ds;
    ds = delta_sol(lsol, obli);
    return(Math.sin(phi) * Math.sin(ds) + Math.cos(phi) * Math.cos(ds) * Math.cos(hsv - Math.PI));
}

function leve(phi, lsol, hsv, obli) {
    return(sinush(phi, lsol, hsv, obli) > 0 ? 1 : 0); // 1 si levé, 0 si couché
}
